import unittest
import os

from gym_mapf.envs.utils import parse_scen_file


class ParsersTest(unittest.TestCase):

    def test_scen_parser_emtpy_8_8(self):
        scen_file_path = os.path.abspath(os.path.join(__file__, '../../maps/empty-8-8/empty-8-8-even-1.scen'))
        agents_starts, agents_goals = parse_scen_file(scen_file_path, 4)

        self.assertEqual(agents_starts, ((0, 0), (5, 3), (1, 7), (0, 5)))
        self.assertEqual(agents_goals, ((1, 0), (5, 6), (6, 4), (7, 4)))


if __name__ == '__main__':
    unittest.main(verbosity=2)



import unittest
import os.path

from gym_mapf.envs.mapf_env import MapfEnv
from gym_mapf.mapf.grid import MapfGrid
from gym_mapf.envs.utils import parse_map_file
from gym_mapf.envs import *
from copy import copy

RIGHT_FAIL = 0.1
LEFT_FAIL = 0.1
REWARD_OF_CLASH = -1000.0
REWARD_OF_LIVING = 0.0
REWARD_OF_GOAL = 100.0


class MapfEnvTest(unittest.TestCase):
    def test_transition_function_empty_grid(self):
        """Assert the basic steps are done right.

        * Define an empty 8x8 environment with two agents starting at (0,0),(7,7) and desire to reach (0,2),(5,7).
        * Perform one (RIGHT, UP) step and assert that the transitions are correct.
        * Perform another (RIGHT, UP) step from the most probable next state from before ((0,1), (6,7)) and assert
            that the transitions are correct again, including the terminal one.
        """
        map_file_path = os.path.abspath(os.path.join(__file__, '../../maps/empty-8-8/empty-8-8.map'))
        grid = MapfGrid(parse_map_file(map_file_path))

        # agents are starting a
        agent_starts, agents_goals = ((0, 0), (7, 7)), ((0, 2), (5, 7))
        env = MapfEnv(grid, agent_starts, agents_goals,
                      RIGHT_FAIL, LEFT_FAIL, REWARD_OF_CLASH, REWARD_OF_GOAL, REWARD_OF_LIVING)

        first_step_transitions = [(round(prob, 2), next_state, reward, done)
                                  for (prob, next_state, reward, done) in env.P[env.s][(RIGHT, UP)]]

        self.assertEqual(set(first_step_transitions), {
            (0.64, ((0, 1), (6, 7)), REWARD_OF_LIVING, False),  # (RIGHT, UP)
            (0.08, ((1, 0), (6, 7)), REWARD_OF_LIVING, False),  # (DOWN, UP)
            (0.08, ((0, 0), (6, 7)), REWARD_OF_LIVING, False),  # (UP, UP)
            (0.08, ((0, 1), (7, 7)), REWARD_OF_LIVING, False),  # (RIGHT, RIGHT)
            (0.08, ((0, 1), (7, 6)), REWARD_OF_LIVING, False),  # (RIGHT, LEFT)
            (0.01, ((1, 0), (7, 7)), REWARD_OF_LIVING, False),  # (DOWN, RIGHT)
            (0.01, ((1, 0), (7, 6)), REWARD_OF_LIVING, False),  # (DOWN, LEFT)
            (0.01, ((0, 0), (7, 7)), REWARD_OF_LIVING, False),  # (UP, RIGHT)
            (0.01, ((0, 0), (7, 6)), REWARD_OF_LIVING, False)  # (UP, LEFT)
        })

        wish_state = ((0, 1), (6, 7))
        second_step_transitions = [(round(prob, 2), next_state, reward, done)
                                   for (prob, next_state, reward, done) in env.P[wish_state][(RIGHT, UP)]]

        # [(0,0), (7,7)]
        self.assertEqual(set(second_step_transitions), {
            (0.64, ((0, 2), (5, 7)), REWARD_OF_GOAL, True),  # (RIGHT, UP)
            (0.08, ((1, 1), (5, 7)), REWARD_OF_LIVING, False),  # (DOWN, UP)
            (0.08, ((0, 1), (5, 7)), REWARD_OF_LIVING, False),  # (UP, UP)
            (0.08, ((0, 2), (6, 7)), REWARD_OF_LIVING, False),  # (RIGHT, RIGHT)
            (0.08, ((0, 2), (6, 6)), REWARD_OF_LIVING, False),  # (RIGHT, LEFT)
            (0.01, ((1, 1), (6, 7)), REWARD_OF_LIVING, False),  # (DOWN, RIGHT)
            (0.01, ((1, 1), (6, 6)), REWARD_OF_LIVING, False),  # (DOWN, LEFT)
            (0.01, ((0, 1), (6, 7)), REWARD_OF_LIVING, False),  # (UP, RIGHT)
            (0.01, ((0, 1), (6, 6)), REWARD_OF_LIVING, False)  # (UP, LEFT)
        })

    def test_colliding_agents_state_is_terminal_and_negative_reward(self):
        map_file_path = os.path.abspath(os.path.join(__file__, '../../maps/empty-8-8/empty-8-8.map'))

        grid = MapfGrid(parse_map_file(map_file_path))

        # agents are starting a
        agent_starts, agents_goals = ((0, 0), (0, 2)), ((7, 7), (5, 5))
        env = MapfEnv(grid, agent_starts, agents_goals,
                      RIGHT_FAIL, LEFT_FAIL, REWARD_OF_CLASH, REWARD_OF_GOAL, REWARD_OF_LIVING)
        transitions = [(round(prob, 2), next_state, reward, done)
                       for (prob, next_state, reward, done) in env.P[env.s][(RIGHT, LEFT)]]

        self.assertIn((0.64, ((0, 1), (0, 1)), REWARD_OF_CLASH, True), set(transitions))

    def test_agent_doesnt_move_if_reach_to_goal(self):
        grid = MapfGrid([
            '....',
            '....',
            '....',
            '....'])

        agent_starts, agents_goals = ((0, 0), (3, 3)), ((0, 0), (1, 3))  # one agent is already at it's goal
        env = MapfEnv(grid, agent_starts, agents_goals,
                      RIGHT_FAIL, LEFT_FAIL, REWARD_OF_CLASH, REWARD_OF_GOAL, REWARD_OF_LIVING)

        transitions = [(round(prob, 2), next_state, reward, done)
                       for (prob, next_state, reward, done) in env.P[env.s][(RIGHT, UP)]]

        self.assertEqual(set(transitions), {
            (0.8, ((0, 0), (2, 3)), REWARD_OF_LIVING, False),  # (STAY, UP)
            (0.1, ((0, 0), (3, 3)), REWARD_OF_LIVING, False),  # (STAY, RIGHT)
            (0.1, ((0, 0), (3, 2)), REWARD_OF_LIVING, False),  # (STAY, LEFT)
        })

    def test_soc_makespan(self):
        grid = MapfGrid([
            '....',
            '....',
            '....',
            '....'])

        agent_starts, agents_goals = ((0, 0), (3, 3), (1, 1)), ((0, 1), (1, 3), (1, 2))
        determinstic_env = MapfEnv(grid, agent_starts, agents_goals,
                                   0.0, 0.0, REWARD_OF_CLASH, REWARD_OF_GOAL, REWARD_OF_LIVING)

        determinstic_env.step((RIGHT, UP, RIGHT))
        s, r, done, _ = determinstic_env.step((RIGHT, UP, RIGHT))

        self.assertEqual(s, ((0, 1), (1, 3), (1, 2)))
        self.assertEqual(r, REWARD_OF_GOAL)
        self.assertEqual(determinstic_env.soc, 4)
        self.assertEqual(determinstic_env.makespan, 2)

    def test_soc_makespan_single_agent(self):
        grid = MapfGrid([
            '....',
            '....',
            '....',
            '....',
            '....'])
        determinstic_env = MapfEnv(grid, ((0, 0),), ((4, 0),),
                                   0.0, 0.0, REWARD_OF_CLASH, REWARD_OF_GOAL, REWARD_OF_LIVING)

        determinstic_env.step((DOWN,))
        determinstic_env.step((DOWN,))
        determinstic_env.step((DOWN,))
        s, r, done, _ = determinstic_env.step((DOWN,))

        self.assertEqual(s, ((4, 0),))
        self.assertEqual(r, REWARD_OF_GOAL)
        self.assertEqual(determinstic_env.soc, 4)
        self.assertEqual(determinstic_env.makespan, 4)

    def test_copy_mapf_env(self):
        grid = MapfGrid([
            '....',
            '....',
            '....',
            '....',
            '....'])
        env = MapfEnv(grid, ((0, 0),), ((4, 0),),
                      0.0, 0.0, REWARD_OF_CLASH, REWARD_OF_GOAL, REWARD_OF_LIVING)

        self.assertEqual(env.soc, 0)
        self.assertEqual(env.makespan, 0)

        env.step((RIGHT,))

        self.assertEqual(env.soc, 1)
        self.assertEqual(env.makespan, 1)

        env_copy = copy(env)
        env_copy.step((RIGHT,))

        self.assertEqual(env.soc, 1)
        self.assertEqual(env.makespan, 1)
        self.assertEqual(env_copy.soc, 2)
        self.assertEqual(env_copy.makespan, 2)

    def test_action_from_terminal_state_has_no_effect(self):
        grid = MapfGrid(['..',
                         '..'])
        env = MapfEnv(grid, ((0, 0),), ((1, 1),),
                      0.0, 0.0, REWARD_OF_CLASH, REWARD_OF_GOAL, REWARD_OF_LIVING)

        state, reward, done, _ = env.step((RIGHT,))
        self.assertEqual(reward, 0)
        self.assertEqual(done, False)
        state, reward, done, _ = env.step((DOWN,))
        self.assertEqual(reward, REWARD_OF_GOAL)
        self.assertEqual(done, True)
        # now, after the game is finished - do another step and make sure it has not effect.
        state_after_done, reward_after_done, done_after_done, _ = env.step((UP,))
        self.assertEqual(state_after_done, state)
        self.assertEqual(done_after_done, True)
        self.assertEqual(reward_after_done, 0)
        # another time like I'm trying to reach the goal
        state_after_done, reward_after_done, done_after_done, _ = env.step((DOWN,))
        self.assertEqual(state_after_done, state)
        self.assertEqual(done_after_done, True)
        self.assertEqual(reward_after_done, 0)


if __name__ == '__main__':
    unittest.main(verbosity=2)



from gym_mapf.envs import map_name_to_files
from gym_mapf.mapf.grid import MapfGrid
from gym_mapf.envs.mapf_env import MapfEnv


def parse_scen_file(scen_file, n_agents):
    """Return the agent start locations and the goal locations.

    Args:
        scen_file (str): path to the scenario file.
        n_agents (int): number of agents to read from the scenario (might contain a lot of agents - the more the harder)

    Returns:
        tuple. two lists - one of start locations and one of goal locations (each locations is a tuple of x,y).
    """
    starts = []
    goals = []
    with open(scen_file, 'r') as f:
        lines = iter(f)
        next(lines)
        for i, line in enumerate(lines):
            _, _, _, _, x_start, y_start, x_goal, y_goal, _ = line.split('\t')
            starts.append((int(x_start), int(y_start)))
            goals.append((int(x_goal), int(y_goal)))
            if i == n_agents - 1:
                break

    return tuple(starts), tuple(goals)


def parse_map_file(map_file):
    with open(map_file, 'r') as f:
        lines = f.readlines()

    return lines[4:]


def create_mapf_env(map_name, scen_id, n_agents, right_fail, left_fail, reward_of_clash, reward_of_goal,
                    reward_of_living):
    map_file, scen_file = map_name_to_files(map_name, scen_id)
    grid = MapfGrid(parse_map_file(map_file))
    agents_starts, agents_goals = parse_scen_file(scen_file, n_agents)

    env = MapfEnv(grid, agents_starts, agents_goals,
                  right_fail, left_fail, reward_of_clash, reward_of_goal, reward_of_living)

    return env





import unittest

from gym_mapf.envs.utils import create_mapf_env
from gym_mapf.envs import integer_to_vector, vector_to_integer
from gym_mapf.envs.mapf_env import ACTIONS, UP, RIGHT, DOWN, LEFT, STAY


class UtilsTest(unittest.TestCase):
    def test_create_mapf_env(self):
        empty_8_8_1 = create_mapf_env(map_name='empty-8-8',
                                      scen_id=1,
                                      n_agents=2,
                                      right_fail=0.1,
                                      left_fail=0.1,
                                      reward_of_clash=-1000.0,
                                      reward_of_goal=100.0,
                                      reward_of_living=0.0)

        self.assertEqual(empty_8_8_1.s, ((0, 0), (5, 3)))

        empty_48_48_16 = create_mapf_env(map_name='empty-48-48',
                                         scen_id=16,
                                         n_agents=2,
                                         right_fail=0.1,
                                         left_fail=0.1,
                                         reward_of_clash=-1000.0,
                                         reward_of_goal=100.0,
                                         reward_of_living=0.0)

        self.assertEqual(empty_48_48_16.s, ((40, 42), (17, 2)))

    def test_integer_to_vector(self):
        # state in a 4x4 grid for a single agent.
        self.assertEqual(integer_to_vector(10, 4, 2, lambda n: n), (2, 2))

        # action for 3 agents
        self.assertEqual(integer_to_vector(28, 5, 3, lambda n: ACTIONS[n]), (LEFT, UP, RIGHT))

        # state in a 4x3 grid for two agents.
        self.assertEqual(integer_to_vector(10, 4 * 3, 2, lambda n: (int(n / 3), n % 3)),
                         ((3, 1), (0, 0)))
        self.assertEqual(integer_to_vector(13, 4 * 3, 2, lambda n: (int(n / 3), n % 3)),
                         ((0, 1), (0, 1)))
        self.assertEqual(integer_to_vector(14, 4 * 3, 2, lambda n: (int(n / 3), n % 3)),
                         ((0, 2), (0, 1)))
        self.assertEqual(integer_to_vector(23, 4 * 3, 2, lambda n: (int(n / 3), n % 3)),
                         ((3, 2), (0, 1)))
        self.assertEqual(integer_to_vector(143, 4 * 3, 2, lambda n: (int(n / 3), n % 3)),
                         ((3, 2), (3, 2)))

    def test_vector_to_integer(self):
        # state in a 4x4 grid for a single agent.
        self.assertEqual(vector_to_integer((2, 2), 4, lambda n: n), 10)

        # action for 3 agents
        self.assertEqual(vector_to_integer((LEFT, UP, RIGHT), 5, lambda a: ACTIONS.index(a)), 28)

        # state in a 4x3 grid for two agents.
        self.assertEqual(vector_to_integer(((3, 1), (0, 0)), 4 * 3, lambda v: 3 * v[0] + v[1]), 10)
        self.assertEqual(vector_to_integer(((0, 1), (0, 1)), 4 * 3, lambda v: 3 * v[0] + v[1]), 13)
        self.assertEqual(vector_to_integer(((0, 2), (0, 1)), 4 * 3, lambda v: 3 * v[0] + v[1]), 14)
        self.assertEqual(vector_to_integer(((3, 2), (0, 1)), 4 * 3, lambda v: 3 * v[0] + v[1]), 23)
        self.assertEqual(vector_to_integer(((3, 2), (3, 2)), 4 * 3, lambda v: 3 * v[0] + v[1]), 143)


if __name__ == '__main__':
    unittest.main(verbosity=2)



from collections import Counter

import numpy as np
from colorama import Fore
from gym import spaces
from gym.envs.toy_text.discrete import DiscreteEnv

from gym_mapf.envs import *
from gym_mapf.mapf.grid import EmptyCell, ObstacleCell

CELL_TO_CHAR = {
    EmptyCell: '.',
    ObstacleCell: '@'
}

np_random = np.random.RandomState()

np_random.seed(0)


def stay_if_hit_obstacle(exec_func):
    def new_exec_func(loc, map):
        new_loc = exec_func(loc, map)
        if map[new_loc] is ObstacleCell:
            return loc

        return new_loc

    return new_exec_func


@stay_if_hit_obstacle
def execute_up(loc, map):
    return max(0, loc[0] - 1), loc[1]


@stay_if_hit_obstacle
def execute_down(loc, map):
    return min(len(map) - 1, loc[0] + 1), loc[1]


@stay_if_hit_obstacle
def execute_right(loc, map):
    return loc[0], min(len(map[0]) - 1, loc[1] + 1)


@stay_if_hit_obstacle
def execute_left(loc, map):
    return loc[0], max(0, loc[1] - 1)


def execute_stay(loc, _):
    return loc


ACTION_TO_FUNC = {
    UP: execute_up,
    DOWN: execute_down,
    RIGHT: execute_right,
    LEFT: execute_left,
    STAY: execute_stay
}


def execute_action(grid, s, noised_action):
    new_state = []
    for i, single_action in enumerate(noised_action):
        exec_func = ACTION_TO_FUNC[single_action]
        new_state.append(exec_func(s[i], grid))

    new_state = tuple(new_state)
    return new_state


class StateToActionGetter:
    def __init__(self, grid, agents_starts, agents_goals,
                 right_fail, left_fail, reward_of_clash, reward_of_goal, reward_of_living,
                 s):
        self.grid = grid
        self.agents_starts = agents_starts
        self.agents_goals = agents_goals
        self.s = s
        self.right_fail = right_fail
        self.left_fail = left_fail
        self.reward_of_clash = reward_of_clash
        self.reward_of_goal = reward_of_goal
        self.reward_of_living = reward_of_living

    def get_possible_actions(self, a):
        if len(a) == 1:
            right, left = POSSIBILITIES[a[0]]
            return [
                (self.right_fail, (right,)),
                (self.left_fail, (left,)),
                (1.0 - self.right_fail - self.left_fail, a)
            ]

        head, *tail = a
        tail = tuple(tail)
        right, left = POSSIBILITIES[head]
        res = []
        for prob, noised_action in self.get_possible_actions(tail):
            res += [
                (self.right_fail * prob, (right,) + noised_action),  # The first action noised to right
                (self.left_fail * prob, (left,) + noised_action),  # The first action noised to left
                ((1.0 - self.right_fail - self.left_fail) * prob, (head,) + noised_action)
                # The first action remained the same
            ]

        return res

    def is_terminal(self, s):
        loc_count = Counter(s)
        if len([x for x in loc_count.values() if x > 1]) != 0:  # clash between two agents.
            return True

        goals = [loc == self.agents_goals[i]
                 for i, loc in enumerate(s)]
        all_in_goal = all(goals)

        if all_in_goal:
            return True

        return False

    def calc_transition_reward(self, original_state, action, new_state):
        loc_count = Counter(new_state)
        if len([x for x in loc_count.values() if x > 1]) != 0:  # clash between two agents.
            return self.reward_of_clash, True

        goals = [loc == self.agents_goals[i]
                 for i, loc in enumerate(new_state)]
        all_in_goal = all(goals)

        if all_in_goal:
            return self.reward_of_goal, True

        return self.reward_of_living, False

    def __getitem__(self, a):
        if type(a) == int:
            # Hidden assumption - if one sent a as an int he also sent the state as an int
            # and want to receive it back as an int.
            a = integer_to_vector(a, len(ACTIONS), len(self.agents_goals), lambda n: ACTIONS[n])
            should_return_int = True
        else:
            should_return_int = False

        if self.is_terminal(self.s):
            if should_return_int:
                return [(1.0, vector_to_integer(self.s, len(self.grid) * len(self.grid[0]),
                                                lambda v: len(self.grid[0]) * v[0] + v[1]), 0, True)]

            else:
                return [(1.0, self.s, 0, True)]

        transitions = []
        for i in range(len(self.agents_starts)):
            if self.s[i] == self.agents_goals[i]:
                a = a[:i] + (STAY,) + a[(i + 1):]
        for prob, noised_action in self.get_possible_actions(a):
            new_state = execute_action(self.grid, self.s, noised_action)
            reward, done = self.calc_transition_reward(self.s, a, new_state)
            similar_transitions = [(p, s, r, d) for (p, s, r, d) in transitions
                                   if s == new_state and r == reward and d == done]
            if len(similar_transitions) > 0:
                idx = transitions.index(similar_transitions[0])
                old_prob, _, _, _ = transitions[idx]
                transitions = transitions[:idx] + [(old_prob + prob, new_state, reward, done)] + transitions[(idx + 1):]
            else:
                transitions.append((prob, new_state, reward, done))

        if should_return_int:
            return [(prob,
                     vector_to_integer(new_state, len(self.grid) * len(self.grid[0]),
                                       lambda v: len(self.grid[0]) * v[0] + v[1]),
                     reward,
                     done)
                    for (prob, new_state, reward, done) in transitions]
        else:
            return transitions


class StateGetter:
    def __init__(self, grid, agents_starts, agent_goals, right_fail, left_fail, reward_of_clash, reward_of_goal,
                 reward_of_living, ):
        self.grid = grid
        self.agents_starts = agents_starts
        self.agents_goals = agent_goals
        self.right_fail = right_fail
        self.left_fail = left_fail
        self.reward_of_clash = reward_of_clash
        self.reward_of_goal = reward_of_goal
        self.reward_of_living = reward_of_living

    def __getitem__(self, s):
        if type(s) == int:
            s = integer_to_vector(s,
                                  len(self.grid) * len(self.grid[0]),
                                  len(self.agents_starts),
                                  lambda n: (int(n / len(self.grid[0])), n % len(self.grid[0])))

        return StateToActionGetter(self.grid,
                                   self.agents_starts,
                                   self.agents_goals,
                                   self.right_fail,
                                   self.left_fail,
                                   self.reward_of_clash,
                                   self.reward_of_goal,
                                   self.reward_of_living,
                                   s)


class SingleActionSpace(spaces.Discrete):
    NUM_TO_ACTION = {
        0: UP,
        1: RIGHT,
        2: DOWN,
        3: LEFT,
        4: STAY
    }

    def __init__(self):
        super().__init__(5)

    def sample(self):
        return self.NUM_TO_ACTION[np_random.randint(self.n)]

    def __repr__(self):
        return "Discrete(%s, %s, %s, %s, %s)" % (UP, RIGHT, DOWN, LEFT, STAY)


class MapfEnv(DiscreteEnv):

    # TODO: return to call super c'tor
    def __init__(self, grid, agents_starts, agents_goals,
                 right_fail, left_fail, reward_of_clash, reward_of_goal, reward_of_living):
        if len(agents_starts) != len(agents_goals):
            raise Exception("Illegal Arguments - agents starts and goals must have the same length")

        self.grid = grid
        self.agents_starts, self.agents_goals = agents_starts, agents_goals
        self.n_agents = len(agents_starts)
        self.right_fail = right_fail
        self.left_fail = left_fail
        self.reward_of_clash = reward_of_clash
        self.reward_of_goal = reward_of_goal
        self.reward_of_living = reward_of_living

        self.nS = (len(self.grid) * len(self.grid[0])) ** self.n_agents  # each agent may be in each of the cells.
        self.nA = len(ACTIONS) ** self.n_agents
        self.P = StateGetter(self.grid, self.agents_starts, agents_goals, right_fail, left_fail,
                             reward_of_clash, reward_of_goal, reward_of_living)
        self.isd = [1.0] + [0.0] * (self.nS - 1)  # irrelevant.
        self.lastaction = None  # for rendering

        self.action_space = spaces.Tuple([SingleActionSpace()] * self.n_agents)
        self.observation_space = spaces.Tuple(
            [spaces.Tuple(
                [spaces.Discrete(len(self.grid)),
                 spaces.Discrete(len(self.grid[0]))])] * self.n_agents)
        self.makespan = 0
        self.soc = 0

        self.seed()
        self.reset()

    def step(self, a):
        n_agents_in_goals = len([i for i in range(self.n_agents) if self.s[i] == self.agents_goals[i]])

        self.soc += self.n_agents - n_agents_in_goals
        self.makespan += 1
        return super().step(a)

    def reset(self):
        self.lastaction = None
        self.makespan = 0
        self.soc = 0
        self.s = self.agents_starts
        return self.s

    def render(self, mode='human'):
        # init(autoreset=True)

        for i in range(len(self.grid)):
            print('')  # newline
            for j in range(len(self.grid[0])):
                if (i, j) in self.s and (i, j) in self.agents_goals:
                    # print an agent which reached it's goal
                    print(Fore.GREEN + str(self.s.index((i, j))) + Fore.RESET, end=' ')
                    continue
                if (i, j) in self.s:
                    print(Fore.YELLOW + str(self.s.index((i, j))) + Fore.RESET, end=' ')
                    continue
                if (i, j) in self.agents_goals:
                    print(Fore.BLUE + str(self.agents_goals.index((i, j))) + Fore.RESET, end=' ')
                    continue

                print(CELL_TO_CHAR[self.grid[i, j]], end=' ')
